
from tqdm.auto import tqdm
import ujson
from pathlib import Path
from . import openalex
import pandas as pd
import csv
import math
import warnings
import numpy as np
# import openalex

def extractEntry(data,path):
    for pathIndex,key in enumerate(path):
        if(isinstance(data,dict)):
            if(key in data):
                data = data[key]
            else:
                return None
        elif(isinstance(data,list)):
            return [extractEntry(entry,path[pathIndex:]) for entry in data]
        else:
            return None
        
    return data

dataTypeMap = { # Default
    "i":("int",None,-1),
    "s":("str",None,""),
    "f":("float",None,np.nan),
    "d":("double",None,np.nan),
    "a":("json",None,""),
    "I":("json","int", -1),
    "F":("json","float", np.nan),
    "D":("json","float", np.nan),
    "S":("json","str", ""),
    "F":("json","int",-1),
}

dataTypeMapStringToPandasType = {
    "int":"Int64",
    "str":"string",
    "float":"float64",
    "double":"float64",
    "json":"string",
}


def processOpenAlexID(entityID):
    if(entityID):
        if(isinstance(entityID,str)):
            return int(entityID.replace("https://openalex.org/","")[1:])
        elif(isinstance(entityID,int)):
            return entityID
        else:
            print("Error: ID is not a string or an integer: ",entityID)
            return -1
    else:
        return -1
    
schemasPerCategory = {}
schemasPerCategory["works"] = {
    "core":[
        ("id", "i", processOpenAlexID),
    ],
    "basic":[
        ("title", "s", None),
        ("type", "s", None),
        ("publication_year", "i", None),
        ("publication_date", "s", None),
        ("doi", "s", None),
        ("primary_location:source:id", "i", processOpenAlexID),
        ("primary_location:source:type", "s", None),
        ("is_retracted", "s", str),
        ("is_paratext", "s", str),
    ],
    "authorship":[
        ("authorships:author_position", "S", None),
        ("authorships:author:id", "I", processOpenAlexID),
        ("authorships:institutions:id", "a", lambda entry: [processOpenAlexID(value) for value in entry] ),
        ("corresponding_author_ids", "I", processOpenAlexID),
        ("corresponding_institution_ids", "I", processOpenAlexID),
    ],
    "ids":[
        ("ids:doi", "s", None),
        ("ids:mag", "i", None),
        ("ids:pmcid", "s", None),
        ("ids:pmid", "s", None),
    ],
    "biblio":[
        ("biblio:volume", "s", None),
        ("biblio:issue", "s", None),
        ("biblio:first_page", "s", None),
        ("biblio:last_page", "s", None),
    ],
    "open_access":[
        ("open_access:is_oa","s", str),
        ("open_access:oa_status", "s", None),
        ("open_access:any_repository_has_fulltext", "s", None),
        ("open_access:oa_status", "s", None),
    ],
    "open_access_location":[
        ('best_oa_location:is_oa',"s", str),
        ('best_oa_location:landing_page_url',"s", None),
        ('best_oa_location:license', "s", None),
        ('best_oa_location:pdf_url', "s", None),
        ('best_oa_location:source:display_name', "s", None),
        ('best_oa_location:source:host_organization', "s", None),
        ('best_oa_location:source:host_organization_lineage', "s", None),
        ('best_oa_location:source:host_organization_name', "s", None),
        ('best_oa_location:source:id', "i", processOpenAlexID),
        ('best_oa_location:source:issn', "s", None),
        ('best_oa_location:source:issn_l', "s", None),
        ('best_oa_location:source:publisher', "s", None),
        ('best_oa_location:source:publisher_id', "i", processOpenAlexID),
        ('best_oa_location:source:type', "s", None),
        ('best_oa_location:version', "s", None),
    ],
    "source":[
        ("primary_location:source:publisher_id", "i", processOpenAlexID),
    ],
    "funding":[
        ("grants:funder", "I", processOpenAlexID),
        ("grants:award_id", "S", str),
    ],
    "concepts":[
        ("concepts:id", "I", processOpenAlexID),
        ("concepts:score", "F", None),
    ],
    "mesh":[
        ("mesh:is_major_topic", "S", str),
        ("mesh:descriptor_ui", "S", None),
        ("mesh:descriptor_name", "S", None),
        ("mesh:qualifier_ui", "S", None),
        ("mesh:qualifier_name", "S", None),
    ],
    "references":[
        ("referenced_works", "I", processOpenAlexID),
    ],
    "abstract":[
        ("abstract_inverted_index", "s", openalex.processInvertedAbstract),
    ],
    "citation_count":[
        ("cited_by_count", "i", None),
    ],
    "citation_count_by_year":[
        ("cited_by_count", "a", None),
    ],
    "apc":[
        ("apc_payment", "a", None),
    ],
    "locations":[
        ("locations", "a", None),
    ]
}

schemasPerCategory["authors"] = {
    "core":[
        ("id", "i", processOpenAlexID),
    ],
    "basic":[
        ("orcid", "s", None),
        ("display_name", "s", None),
        ("last_known_institution:id", "i", processOpenAlexID)
    ],
    "name_alternatives":[
        ("display_name_alternatives", "S", None),
    ],
    "ids":[
        ("ids:orcid", "s", None),
        ("ids:mag", "s", None),
        ("ids:twitter", "s", None),
        ("ids:wikipedia", "s", None),
        ("ids:scopus", "s", None),
    ],
    "citation_count":[
        ("cited_by_count", "i", None),
    ],
    "citation_count_by_year":[
        ("cited_by_count", "a", None),
    ],
    "works_count":[
        ("works_count", "i", None),
    ],
    "concepts":[
        ("x_concepts:id", "I", processOpenAlexID),
        ("x_concepts:score", "F", None),
    ],
    "metrics":[
        ("summary_stats:2yr_h_index", "i", None),
        ("summary_stats:2yr_i10_index", "i", None),
        ("summary_stats:2yr_mean_citedness", "i", None),
        ("summary_stats:h_index", "i", None),
        ("summary_stats:i10_index", "i", None),
    ]
}

# [('abbreviated_title', 0.031994724297111726),
#  ('alternate_titles', 0.24632446434805846),
#  ('apc_prices', 0.06609714362066661),
#  ('apc_usd', 0.02368031999034512),
#  ('cited_by_count', 0.7791790764933817),
#  ('country_code', 0.47910623386363344),
#  ('counts_by_year', 0.9129207308400178),
#  ('created_date', 1.0),
#  ('display_name', 1.0),
#  ('homepage_url', 0.3096070377186895),
#  ('host_organization', 0.29923666096281576),
#  ('host_organization_lineage', 0.2986720228268975),
#  ('host_organization_lineage_names', 0.2986720228268975),
#  ('host_organization_name', 0.29923666096281576),
#  ('id', 1.0),
#  ('ids:fatcat', 0.15712456951729906),
#  ('ids:issn', 0.5118724866060076),
#  ('ids:issn_l', 0.5118724866060076),
#  ('ids:mag', 0.21006262742072437),
#  ('ids:openalex', 1.0),
#  ('ids:wikidata', 0.373570625024245),
#  ('is_in_doaj', 0.06198519872245234),
#  ('is_oa', 0.17898166865654916),
#  ('issn', 0.5118724866060076),
#  ('issn_l', 0.5118724866060076),
#  ('publisher', 0.5937881184619429),
#  ('publisher_id', 0.2937109656174167),
#  ('societies', 0.016689151620425245),
#  ('summary_stats:2yr_cited_by_count', 0.5922666126453081),
#  ('summary_stats:2yr_h_index', 0.28759477084743135),
#  ('summary_stats:2yr_i10_index', 0.07492015327123751),
#  ('summary_stats:2yr_mean_citedness', 0.288823182059162),
#  ('summary_stats:2yr_works_count', 0.4546931773610279),
#  ('summary_stats:cited_by_count', 0.7791790764933817),
#  ('summary_stats:h_index', 0.7901270220295077),
#  ('summary_stats:i10_index', 0.3740016465020452),
#  ('summary_stats:oa_percent', 0.45804652445831373),
#  ('summary_stats:works_count', 0.9480920834285173),
#  ('type', 1.0),
#  ('updated_date', 1.0),
#  ('works_api_url', 1.0),
#  ('works_count', 0.9480920834285173),
#  ('x_concepts', 0.9258298241001349)]
schemasPerCategory["sources"] = {
    "core":[
        ("id", "i", processOpenAlexID),
    ],
    "basic":[
        ("display_name", "s", None),
        ("type", "s", None),
        ("issn", "S", None),
        ("issn_l", "s", None),
        ("publisher", "s", None),
        ("publisher_id", "i", processOpenAlexID),
    ],
    "open_access":[
        ("is_oa", "s", None),
        ("is_in_doaj", "s", None),
    ],
    "urls":[
        ("homepage_url", "s", None),
    ],
    "ids":[
        ("ids:issn", "S", None),
        ("ids:issn_l", "s", None),
        ("ids:mag", "i", processOpenAlexID),
        ("ids:wikidata", "s", None),
        ("ids:fatcat", "s", None),
    ],
    "metrics":[
        ("summary_stats:2yr_cited_by_count", "i", None),
        ("summary_stats:2yr_h_index", "i", None),
        ("summary_stats:2yr_i10_index", "i", None),
        ("summary_stats:2yr_mean_citedness", "i", None),
        ("summary_stats:2yr_works_count", "i", None),
        ("summary_stats:cited_by_count", "i", None),
        ("summary_stats:h_index", "i", None),
        ("summary_stats:i10_index", "i", None),
    ],
    "citation_count":[
        ("cited_by_count", "i", None),
    ],
    "works_count":[
        ("works_count", "i", None),
    ],
    "country":[
        ("country_code", "s", None),
    ],
    "name_alternatives":[
        ("alternate_names", "S", None),
    ],
    "societies":[
        ("societies", "a", None),
    ],
    "concepts":[
        ("x_concepts:id", "I", processOpenAlexID),
        ("x_concepts:score", "F", None),
    ]
}

# [('associated_institutions', 0.19774776970701508),
#  ('cited_by_count', 0.8456783503144347),
#  ('country_code', 0.9999122507678058),
#  ('counts_by_year', 0.8731536099059133),
#  ('created_date', 1.0),
#  ('display_name', 1.0),
#  ('display_name_acronyms', 0.39979525179154685),
#  ('display_name_alternatives', 0.20008774923219422),
#  ('geo:city', 0.9983132647589333),
#  ('geo:country', 0.9983132647589333),
#  ('geo:country_code', 0.9999122507678058),
#  ('geo:geonames_city_id', 0.9980110174035978),
#  ('geo:latitude', 0.993789304343587),
#  ('geo:longitude', 0.993789304343587),
#  ('geo:region', 0.36350606932189344),
#  ('homepage_url', 0.9848778823185297),
#  ('id', 1.0),
#  ('ids:grid', 0.9997855018768586),
#  ('ids:mag', 0.2045532101594111),
#  ('ids:openalex', 1.0),
#  ('ids:ror', 1.0),
#  ('ids:wikidata', 0.2005459952225418),
#  ('ids:wikipedia', 0.34800370496758154),
#  ('image_thumbnail_url', 0.24326037147174961),
#  ('image_url', 0.24326037147174961),
#  ('relationship', 1.9499829376492957e-05),
#  ('repositories', 0.017374347974455223),
#  ('roles', 1.0),
#  ('ror', 1.0),
#  ('summary_stats:2yr_cited_by_count', 0.8018622337054551),
#  ('summary_stats:2yr_h_index', 0.6154926144396237),
#  ('summary_stats:2yr_i10_index', 0.40727343635743185),
#  ('summary_stats:2yr_mean_citedness', 0.6051869546141472),
#  ('summary_stats:2yr_works_count', 0.6977038950909179),
#  ('summary_stats:cited_by_count', 0.8456783503144347),
#  ('summary_stats:h_index', 0.8365816799103007),
#  ('summary_stats:i10_index', 0.7559596353531907),
#  ('summary_stats:oa_percent', 0.7925998147516209),
#  ('summary_stats:works_count', 0.8901769609515917),
#  ('type', 0.999483254521523),
#  ('updated_date', 1.0),
#  ('works_api_url', 1.0),
#  ('works_count', 0.8901769609515917),
#  ('x_concepts', 0.8901379612928387)]
schemasPerCategory["institutions"] = {
    "core":[
        ("id", "i", processOpenAlexID),
    ],
    "basic":[
        ("display_name", "s", None),
        ("type", "s", None),
        ("country_code", "s", None),
    ],
    "alternate_names":[
        ("display_name_acronyms", "S", None),
        ("display_name_alternatives", "S", None),
    ],
    "ids":[
        ("ids:grid", "s", None),
        ("ids:mag", "s", None),
        ("ids:wikidata", "s", None),
        ("ids:wikipedia", "s", None),
        ("ids:ror", "s", None),
    ],
    "geo":[
        ("geo:city", "s", None),
        ("geo:geonames_city_id", "s", None),
        ("geo:region", "s", None),
        ("geo:country_code", "s", None),
        ("geo:country", "s", None),
        ("geo:latitude", "f", None),
        ("geo:longitude", "f", None),
    ],
    "urls":[
        ("homepage_url", "s", None),
        ("image_url", "s", None),
        ("image_thumbnail_url", "s", None),
    ],
    "associated_institutions":[
        ("associated_institutions:id", "I", processOpenAlexID),
        ("associated_institutions:relationship", "S", None),
    ],
    "concepts":[
        ("x_concepts:id", "I", processOpenAlexID),
        ("x_concepts:score", "F", None),
    ],
    "metrics":[
        ("summary_stats:2yr_cited_by_count", "i", None),
        ("summary_stats:2yr_h_index", "i", None),
        ("summary_stats:2yr_i10_index", "i", None),
        ("summary_stats:2yr_mean_citedness", "f", None),
        ("summary_stats:2yr_works_count", "i", None),
        ("summary_stats:cited_by_count", "i", None),
        ("summary_stats:h_index", "i", None),
        ("summary_stats:i10_index", "i", None),
        ("summary_stats:oa_percent", "f", None),
        ("summary_stats:works_count", "i", None),
    ],
    "counts_by_year":[
        ("counts_by_year", "a", None),
    ],
    "roles":[
        ("roles", "a", None),
    ]
    # TODO: Add missing entries
}

schemasPerCategory["concepts"] = {
    "core":[
        ("id", "i", processOpenAlexID),
    ],
    "basic":[
        ("display_name", "s", None),
        ("level", "i", None),
        ("description", "s", None),
        ("ancestors:id", "I", processOpenAlexID),
        ("related_concepts:id", "I", processOpenAlexID),
    ],
    # ("wikidata", "s", None),
    # ("image_url", "s", None),
    # ("image_thumbnail_url", "s", None),
    # TODO: add missing entries
}

# [('alternate_titles', 0.12709760699036837),
#  ('cited_by_count', 0.9680270082414855),
#  ('country_codes', 0.7995233839737861),
#  ('counts_by_year', 0.9796445238804488),
#  ('created_date', 1.0),
#  ('display_name', 1.0),
#  ('hierarchy_level', 0.029490616621983913),
#  ('homepage_url', 0.5567470956210903),
#  ('id', 1.0),
#  ('ids:openalex', 1.0),
#  ('ids:ror', 0.3356171184589415),
#  ('ids:wikidata', 0.8425181213384967),
#  ('image_thumbnail_url', 0.2609472743521001),
#  ('image_url', 0.2609472743521001),
#  ('lineage', 1.0),
#  ('parent_publisher:display_name', 0.029490616621983913),
#  ('parent_publisher:id', 0.029490616621983913),
#  ('roles', 1.0),
#  ('sources_api_url', 1.0),
#  ('sources_count', 0.9945387746996326),
#  ('summary_stats:2yr_cited_by_count', 0.885413563697746),
#  ('summary_stats:2yr_h_index', 0.5894151524178334),
#  ('summary_stats:2yr_i10_index', 0.16284380895640949),
#  ('summary_stats:2yr_mean_citedness', 0.5924932975871313),
#  ('summary_stats:2yr_works_count', 0.747492801112104),
#  ('summary_stats:cited_by_count', 0.9680270082414855),
#  ('summary_stats:h_index', 0.9723959884817793),
#  ('summary_stats:i10_index', 0.6978452983814915),
#  ('summary_stats:oa_percent', 0.6902988779664383),
#  ('summary_stats:sources_count', 0.9945387746996326),
#  ('summary_stats:works_count', 0.9921556945685632),
#  ('updated_date', 1.0),
#  ('works_count', 0.9921556945685632),
#  ('x_concepts', 0.9910634495084897)]
schemasPerCategory["publishers"] = {
    "core":[
        ("id", "i", processOpenAlexID),
    ],
    "basic":[
        ("display_name", "s", None),
        ("country_codes", "s", None),
        ("hierarchy_level", "i", None),
        ("lineage", "I", processOpenAlexID),
    ],
    # TODO: add missing entries
}

# [('alternate_titles', 0.8879674445848876),
#  ('cited_by_count', 0.8848228874433517),
#  ('country_code', 0.9941116626075162),
#  ('counts_by_year', 0.9124764928939174),
#  ('created_date', 1.0),
#  ('description', 0.5320467367512408),
#  ('display_name', 1.0),
#  ('grants_count', 0.7614452631254431),
#  ('homepage_url', 0.5949070505903752),
#  ('id', 1.0),
#  ('ids:openalex', 1.0),
#  ('ids:ror', 0.5425902518728613),
#  ('ids:wikidata', 0.5799241606807042),
#  ('image_thumbnail_url', 0.21728273268181397),
#  ('image_url', 0.21728273268181397),
#  ('roles', 1.0),
#  ('summary_stats:2yr_cited_by_count', 0.8596664303110645),
#  ('summary_stats:2yr_h_index', 0.7630483706877949),
#  ('summary_stats:2yr_i10_index', 0.4962851065141659),
#  ('summary_stats:2yr_mean_citedness', 0.7335758547337916),
#  ('summary_stats:2yr_works_count', 0.8012763202515646),
#  ('summary_stats:cited_by_count', 0.8848228874433517),
#  ('summary_stats:h_index', 0.8777630483706877),
#  ('summary_stats:i10_index', 0.7633566606036316),
#  ('summary_stats:oa_percent', 0.8493078891389463),
#  ('summary_stats:works_count', 0.9124456639023337),
#  ('updated_date', 1.0),
#  ('works_count', 0.9124456639023337),
#  ('x_concepts', 0.9090544748281284)]
schemasPerCategory["funders"] = {
    "core":[
        ("id", "i", processOpenAlexID),
    ],
    "basic":[
        ("display_name", "s", None),
        ("country_code", "s", None),
    ]
    # TODO: add missing entries
}





def createDBGZ(oa,entityType, outputLocation, selection=["core","basic"],filterFunction=None):
    import dbgz
    # oa = openalex.OpenAlex(
    #     openAlexPath = "/gpfs/sciencegenome/OpenAlex/openalex-snapshot"
    # )
    # entityType = "concepts"
    # selection = ["core","basic"]
    # outputLocation = "./Processed/"
    filePrefix = "%s_%s"%(entityType,"+".join(selection))
    archiveName = "%s.dbgz"%filePrefix
    archiveLocation = Path(outputLocation)/(archiveName)
    schemaData = [entry for selectedKey in selection for entry in schemasPerCategory[entityType][selectedKey]]
    schema = [(key,dataType) for key,dataType,postProcess in schemaData]
    entitiesCount = oa.getRawEntityCount(entityType)
    with dbgz.DBGZWriter(archiveLocation, schema) as fdbgz:
        for entity in tqdm(oa.rawEntities(entityType),total=entitiesCount,desc=entityType):
            if(filterFunction is not None):
                if(not filterFunction(entity)):
                    continue
            processedEntity = {}
            for key,dataType,postProcess in schemaData:
                path = key.split(":")
                value = extractEntry(entity,path)
                if(postProcess is not None):
                    if(isinstance(value,list)):
                        value = [postProcess(entry) for entry in value]
                    else:
                        if(dataType!=dataType.lower()): # force list
                            value = [postProcess(value)]
                        else:
                            value = postProcess(value)
                processedEntity[key] = value
            # try:
            # print(processedEntity)
            fdbgz.write(**processedEntity)
            # except AttributeError as e:
            #     print(e)
            #     print(processedEntity)
            #     print("ERROR!!!!")
            #     break

    # entitiesCount = oa.getRawEntityCount(entityType)
    # print(entitiesCount)
    # archiveLocation = "./Processed/%s_simple.dbgz"%entityType
    # indexLocation = "./Processed/%s_simple_byID.idbgz"%entityType
    # print("Saving the index dictionary")
    # with dbgz.DBGZReader(archiveLocation) as fd:
    #     print(fd.entriesCount)
    #     fd.generateIndex(key="id",
    #                     indicesPath=indexLocation,
    #                     useDictionary=False,
    #                     showProgressbar=True
    #                     )

    # Loading a dbgz file
    print("Testing the dbgz file")
    with dbgz.DBGZReader(archiveLocation) as fd:
        entriesCount = fd.entriesCount
        print("\t Number of entries: ", entriesCount)
        for entry in tqdm(fd.entries,total=entriesCount):
            pass




def createTSV(oa,entityType, outputLocation, selection=["core","basic"],filterFunction=None):
    # # Open your tsv file in write mode ('w').
    # oa = openalex.OpenAlex(
    #     openAlexPath = "/gpfs/sciencegenome/OpenAlex/openalex-snapshot"
    # )
    # entityType = "concepts"
    # selection = ["core","basic"]
    # outputLocation = "./Processed/"
    filePrefix = "%s_%s"%(entityType,"+".join(selection))
    archiveName = "%s.tsv"%filePrefix
    archiveJSONMetadataName = "%s.json"%filePrefix
    archiveLocation = Path(outputLocation)/(archiveName)
    schemaData = [entry for selectedKey in selection for entry in schemasPerCategory[entityType][selectedKey]]
    schema = [(key,dataType) for key,dataType,postProcess in schemaData]
    entitiesCount = oa.getRawEntityCount(entityType)
    writtenEntriesCount = 0
    with open(archiveLocation, 'w', newline='') as tsv_file:
        # Use the csv.writer function, but specify a tab ('\t') as the delimiter. Disable quoting and scaping entirely.
        writer = csv.writer(tsv_file, delimiter='\t', quoting=csv.QUOTE_NONE, doublequote=False, escapechar='\\')

        keysInOrder = [key for key,_,_ in schemaData]
        writer.writerow(keysInOrder)

        for entity in tqdm(oa.rawEntities(entityType),total=entitiesCount,desc=entityType):
            processedEntity = {}
            if(filterFunction is not None):
                if(not filterFunction(entity)):
                    continue
            for key,dataType,postProcess in schemaData:
                path = key.split(":")
                value = extractEntry(entity,path)
                _,_,defaultNAValue = dataTypeMap[dataType]
                if(postProcess is not None):
                    if(isinstance(value,list)):
                        value = [postProcess(entry) if entry is not None else defaultNAValue for entry in value if value]
                    else:
                        if(dataType!=dataType.lower()): # force list
                            processedValue = postProcess(value)
                            value = []
                            if(processedValue is not None):
                                value.append(processedValue)
                        else:
                            value = postProcess(value)
                if(dataType!=dataType.lower() or dataType=="a"): # needs json
                    value = ujson.dumps(value)
                if(value is None):
                    value = defaultNAValue
                processedEntity[key] = str(value).replace("\n"," ").replace("\r"," ").replace("\t"," ")
            for key,dataType,postProcess in schemaData:
                _,_,defaultNAValue = dataTypeMap[dataType]
                if(key not in processedEntity or processedEntity[key] is None):
                    if(dataType!=dataType.lower()):
                        processedEntity[key] = str(defaultNAValue)
                    else:
                        processedEntity[key] = str([])
            writer.writerow([processedEntity[key] for key in keysInOrder])
            writtenEntriesCount += 1
        
    with open(Path(outputLocation)/(archiveJSONMetadataName), 'w', newline='') as json_file:
        jsonObject = {}
        jsonSchemaData = {}
        for key,dataType,postProcess in schemaData:
            jsonSchemaData[key] = {}
            dataTypeString,dataSubTypeString,defaultNAValue = dataTypeMap[dataType]
            jsonSchemaData[key]["type"] = dataTypeString
            jsonSchemaData[key]["NA value"] = defaultNAValue
            if(dataSubTypeString is not None):
                jsonSchemaData[key]["subtype"] = dataSubTypeString
        jsonObject["count"] = writtenEntriesCount
        jsonObject["schema"] = jsonSchemaData
        ujson.dump(jsonObject,json_file,indent=4)

            
        # import pandas as pd
        # chunksize = 5000
        # length = 65073
        # with tqdm(total=length, desc="chunks read: ") as bar:
        #     for i, chunk in enumerate(pd.read_csv(archiveLocation, chunksize=chunksize,sep="\t",encoding='utf-8',error_bad_lines=False, low_memory=False,
        #     quoting=csv.QUOTE_NONE)):
        #         bar.update(chunksize)
        # print("Done!!!")
        # print(archiveLocation)


class _chunkIterator:
    """
    Iterator that iterates over all the chunks.
    """
    def __init__(self, chunks, chunksCount):
        self._chunks = chunks
        self._chunkCount = chunksCount


    def __iter__(self):
        for chunk in self._chunks:
            yield chunk
    
    def __len__(self):
        return self._chunkCount
    
class _entryIterator:
    """
    Iterator that iterates over all the entries.
    """
    def __init__(self, chunks, entriesCount):
        self._chunks = chunks
        self._entryCount = entriesCount


    def __iter__(self):
        for chunk in self._chunks:
            for entry in chunk.to_dict(orient="records"):
                yield entry
    
    def __len__(self):
        return self._entryCount

class readTSV:
    def __init__(self,filesLocation, entityType, selection=["core","basic"], chunksize = 5000, convertJSON=True):
        # if filesLocation is a .tsv file, no need to use entityType or selection
        fileLocation = Path(filesLocation)
        if(fileLocation.suffix != ".tsv"):
            filePrefix = "%s_%s"%(entityType,"+".join(selection))
            archiveName = "%s.tsv"%filePrefix
            fileLocation = Path(filesLocation)/(archiveName)
        self.fileLocation = fileLocation
        self.chunksize = chunksize
        self.convertJSON = convertJSON
        #check if metadata json exists
        metadataFileLocation = fileLocation.with_suffix(".json")
        if(metadataFileLocation.exists()):
            with open(metadataFileLocation,"r") as metadataFile:
                self.metadata = ujson.load(metadataFile)
            
            self.archiveDTypes = {}
            for key in self.metadata["schema"]:
                dataTypeString = self.metadata["schema"][key]["type"]
                dataType = dataTypeMapStringToPandasType[dataTypeString]
                self.archiveDTypes[key] = dataType
            self.entriesCount = self.metadata["count"]
            self.chunksCount = math.ceil(self.entriesCount/self.chunksize)

        else:
            self.metadata = None
            self.archiveDTypes = None
            self.entriesCount = -1
            self.chunksCount = -1
            warnings.warn("No metadata file found for %s, will use default dtypes and counter not available."%fileLocation)

    def chunkData(self):
        chunks = pd.read_csv(self.fileHandler, chunksize=self.chunksize, sep="\t", encoding='utf-8',
                                            dtype=self.archiveDTypes,
                                            na_values="None", doublequote=False,escapechar="\\",
                                            on_bad_lines="warn", low_memory=False, quoting=csv.QUOTE_MINIMAL)
        return chunks
    
        #if metadata and types are json, apply json.l
    def __enter__(self):
        # ttysetattr etc goes here before opening and returning the file object
        self.fileHandler = open(self.fileLocation, 'r', newline='')
        return self

    def __exit__(self, type, value, traceback):
        self.fileHandler.close()
        # Exception handling here
        pass

    def __iter__(self):
        for entry in _entryIterator(self.chunkData(),self.entriesCount):
            yield entry
    
    def __len__(self):
        return self.entriesCount
    
    @property
    def entries(self):
        return _entryIterator(self.chunkData(),self.entriesCount)
    
    @property
    def chunks(self):
        return _chunkIterator(self.chunkData(),self.chunksCount)





import warnings
from pathlib import Path
import sys
import os
import ujson
import tempfile
import datetime
# import json_lines SLOWER
import gzip

from collections import OrderedDict
from collections import Counter
import collections


class OpenAlex:
    '''
    OpenAlex is a class for accessing the OpenAlex data snapshots
    '''
    # init

    def __init__(self,
                 openAlexPath = None,
                 processedPath = None,
                 tempPath=None,
                 tagLabel=None,  # current yyyy-mm-dd-hh-mm-ss will be used instead
                 verbose=False
                 ):
        """
        OpenAlex is a class for accessing the OpenAlex data snapshots.

        Parameters
        ----------
        openAlexPath : str or pathlib.Path
            The path to the OpenAlex directory. (default: current working directory)
        processedPath : str or pathlib.Path
            The path to the processed directory used to create indices.
            (default: current working directory, optional if reading from raw data)
        tempPath : str or pathlib.Path
            The path to the temp directory. (default: regular system temp directory)
        tagLabel : str
            The tag label for the current snapshot. (default: current date and time)
        verbose : bool
            If True, print out more information. (default: False)
        """

        if(openAlexPath is None):
            openAlexPath = Path.cwd()

        if (processedPath is None):
            processedPath = Path.cwd()

        if (tempPath is not None):
            tempPath = Path(tempPath)

        if (tagLabel is None):
            tagLabel = self.generateTagLabel()

        self._openAlexPath = Path(openAlexPath)
        self._processedPath = Path(processedPath)
        self._tempPath = tempPath
        self._tagLabel = tagLabel
        self._verbose = verbose

    def getManifestPath(self, entityType):
        """
        Get the path to the manifest file for the given entity type.

        Parameters
        ----------
        entityType : str
            The data: ("authors" | "concepts" | "institutions" | "venues" | "works")


        Returns
        -------
        path
            The path to the manifest file for the given entity type.

        """
        return self._openAlexPath/"data"/entityType/"manifest"

    def getManifest(self, entityType):
        """
        Get the manifest for the given entity type.

        Parameters
        ----------
        entityType : str
            The data: ("authors" | "concepts" | "institutions" | "venues" | "works")


        Returns
        -------
        manifest dict
            The manifest for the given entity type.
        """
        manifestPath = self.getManifestPath(entityType)
        with open(manifestPath, "r") as f:
            manifest = ujson.load(f)
        return manifest

    def getRawEntityCount(self, entityType):
        """
        Get the number of raw entities of the given entity type.

        Parameters
        ----------
        entityType : str
            The data: ("authors" | "concepts" | "institutions" | "venues" | "works")


        Returns
        -------
        int
            The number of raw entities of the given entity type.
        """
        manifestData = self.getManifest(entityType)
        # count = 0
        # for entry in manifestData["entries"]:
        #     count+=int(entry["meta"]["record_count"])
        return manifestData["meta"]["record_count"]

    def getRawEntitiesPaths(self, entityType):
        """
        Get the paths to the raw data on entities of the given entity type.

        Parameters
        ----------
        entityType : str
            The data: ("authors" | "concepts" | "institutions" | "venues" | "works")


        Returns
        -------
        paths iterable
            A iterable collection of paths to the raw data on entities of the given entity type.
        """
        return sorted(self._openAlexPath.glob("data/%s/*/*.gz" % entityType))

    def rawEntities(self, entityType):
        """
        Iterate over the entities of the selected type directly from the raw snapshot.

        Parameters
        ----------
        entityType : str
            The entity type: ("authors" | "concepts" | "institutions" | "venues" | "works")


        Returns
        -------
        entities iterable
            An iterable collection of entities of the selected type.
        """
        from tqdm.auto import tqdm
        for jlpath in self.getRawEntitiesPaths(entityType):
            # SLOWER
            # with json_lines.open(jlpath) as f:
            #     print("starting to read %s"%jlpath)
            #     for item in tqdm(f,total=self.getRawEntityCount(entityType)):
            #          pbar.update(1)
            with gzip.open(jlpath, "rt") as f:
                lineNumber = 0
                for line in f:
                    lineNumber+=1;
                    try:
                        item = ujson.loads(line)
                        yield item
                    except ValueError:
                        warnings.warn("Found problem loading file %s:%d, Content: %s"%(jlpath, lineNumber, str(line)), category=Warning)

    def generateTagLabel(self):
        """
        Generate a tag name for the current processed data based on the
        format yyyy-mm-dd-hh-mm-ss.
        """
        return datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

    def getSchemaForEntityType(self, entityType, reportPath=None, mostCommon=5,
                               saveEach=100000, minPercentageFilter=0, file=sys.stdout):
        """
        Get the schema for the given entity type.

        Parameters
        ----------
        entityType : str
            The entity type: ("authors" | "concepts" | "institutions" | "venues" | "works")
        reportPath : str or pathlib.Path
            The path to save a report file. If not specified, the text report will
            not be sabed. (default: None)
        mostCommon : int
            The number of most common values to be included in the schema. (default: 5)
        saveEach : int
            The number of values to be included in the schema before saving to disk. (default: 1000)

        Returns
        -------
        schema dict
            The schema for the given entity type.
        """
        try:
            from tqdm.auto import tqdm
        except ImportError:
            def tqdm(x): return x

        def saveReport():
            if(reportPath is not None):
                with open(reportPath, "w") as file:
                    _printSchema(schemaData,
                                 file=file,
                                 mostCommon=mostCommon,
                                 minPercentageFilter=minPercentageFilter)

        # authors  concepts  institutions  venues  works
        schemaData = OrderedDict()
        entityCount = 0
        for entity in tqdm(self.rawEntities(entityType), total=self.getRawEntityCount(entityType)):
            _addToSchema(schemaData, entity)
            if(entityCount > 0 and entityCount % saveEach == 0):
                saveReport()
            entityCount += 1
        saveReport()
        return schemaData


def cleanScheme(schemaEntry):
    for key, value in schemaEntry.items():
        if(key == "$count"):
            continue
        if("$samples" in value):
            del value["$samples"]
        if("$listSamples" in value):
            del value["$listSamples"]
        if("$listSamples" in value):
            del value["$listSamples"]
        if("$schemaEntry" in value):
            if(value["$schemaEntry"]):
                cleanScheme(value["$schemaEntry"])
            else:
                del value["$schemaEntry"]

def basicTypeName(obj):
    if(isinstance(obj, str)):
        return "string"
    elif(isinstance(obj, (int, float, complex)) and not isinstance(obj, bool)):
        return "number"
    elif(isinstance(obj, bool)):
        return "bool"
    elif(isinstance(obj, list)):
        return "list"
    else:
        return type(obj).__name__

# from https://github.com/smierz/openalex-utils/blob/8e423319987a07fc7b5599b1b4f19069510d6e9a/notebooks/inverted_index_to_text.ipynb
def processInvertedAbstract(entry):
    invertedAbstract = entry or {}
    abstract_index = {}
    for k, vlist in invertedAbstract.items():
        for v in vlist:
            abstract_index[v] = k
    abstract = ' '.join(abstract_index[k] for k in sorted(abstract_index.keys()))
    return abstract.replace("\n", " ").replace("\r", " ").replace("\t", " ")

def _flatten(d, parent_key=""):
    items = []
    for k, v in d.items():
        if(k == "abstract_inverted_index" and v):
            v = "<ABSTRACT>" # avoid large abstracts in the schema when not null
        k = k.replace(":", "_") # no ":" allowed in keys
        new_key = parent_key+":"+str(k) if parent_key else str(k)
        if isinstance(v, collections.MutableMapping):
            items.extend(_flatten(v, new_key).items())
        else:
            items.append((new_key, v))
    return dict(items)



def _addToSchema(schemaData, entryData, flatDict=True, maxSamples=100):
    if(flatDict):
        entryData = _flatten(entryData)
    if "$count" not in schemaData:
        schemaData["$count"] = 0
    schemaData["$count"] += 1
    for key, value in entryData.items():
        if(key not in schemaData):
            schemaData[key] = {
                "$count": 0,
                "$samples": Counter(),
                "$listSamples": Counter(),
                "$schemaEntry": OrderedDict(),
                "$types": Counter()
            }
        if not value:
            continue
        schemaData[key]["$count"] += 1
        schemaData[key]["$types"][basicTypeName(value)] += 1
        if(not isinstance(value, list)):
            if(len(schemaData[key]["$samples"]) < maxSamples):
                schemaData[key]["$samples"][value] += 1
            else:
                if(value in schemaData[key]["$samples"]):
                    schemaData[key]["$samples"][value] += 1
        else:
            for entry in value:
                if(isinstance(entry, dict)):
                    _addToSchema(
                        schemaData[key]["$schemaEntry"], entry, flatDict=flatDict, maxSamples=maxSamples)
                else:
                    if(len(schemaData[key]["$listSamples"]) < maxSamples):
                        schemaData[key]["$listSamples"][entry] += 1
                    else:
                        if(entry in schemaData[key]["$listSamples"]):
                            schemaData[key]["$listSamples"][entry] += 1




def _percentageText(percentage):
    percentageText = "%.2f%%" % percentage
    if(percentage<0.01):
        percentageText = " < 0.01%"
    return percentageText

def _printSchema(schemaEntry, indent=0, mostCommon=5, minPercentageFilter=0, file=sys.stdout):
    for key, value in schemaEntry.items():
        if(key == "$count"):
            print(" "*indent + str(value)+" ENTRIES", file=file)
            continue
        entriesCount = schemaEntry["$count"]
        if(value["$count"]/entriesCount*100 < minPercentageFilter):
            continue
        # if key instance of tuple
        keyLabel = key
        if(isinstance(key, tuple)):
            keyLabel = ":".join(key)
        print((" "*indent)+keyLabel, file=file)
        print((" "*(indent+2))+"Count: %d (%s)" %
              (value["$count"], _percentageText(value["$count"]/entriesCount*100)), file=file)
        typesStringArray = []
        for sampleItem, sampleCount in value["$types"].most_common(mostCommon):
            typesStringArray.append("%s (%s)" % (
                sampleItem,
                _percentageText(sampleCount/value["$count"]*100)))
        if(len(value["$types"]) > mostCommon):
            typesStringArray.append("...")
        print((" "*(indent+2))+"Types: "+", ".join(typesStringArray), file=file)
        if(value["$samples"]):
            print(" "*(indent+2)+"Samples:", file=file)
            for sampleItem, sampleCount in value["$samples"].most_common(mostCommon):
                print((" "*(indent+4))+"%d (%s): %s" % (sampleCount,
                      _percentageText(sampleCount/value["$count"]*100), sampleItem), file=file)
            if(len(value["$samples"]) > mostCommon):
                print((" "*(indent+4))+"...", file=file)
        if(value["$listSamples"]):
            print(" "*(indent+2)+"List Samples:", file=file)
            for sampleItem, sampleCount in value["$listSamples"].most_common(mostCommon):
                print((" "*(indent+4))+"%d (%s): %s" % (sampleCount,
                      _percentageText(sampleCount/value["$count"]*100), sampleItem), file=file)
            if(len(value["$listSamples"]) > mostCommon):
                print((" "*(indent+4))+"...", file=file)
        if(value["$schemaEntry"]):
            print(" "*(indent+2)+"Schema Entry:", file=file)
            _printSchema(value["$schemaEntry"], indent +
                         4, mostCommon=5, file=file)






import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
   requirements = fh.readlines()

setuptools.setup(
    name="openalex-raw",
    version="0.2.0",
    author="Filipi N. Silva",
    author_email="filipi@iu.edu",
    description="Python library to access OpenAlex Snapshot files",
    install_requires=[req for req in requirements if req[:2] != "# "],
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/filipinascimento/openalex-raw",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5',
)